# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_utils.fixture import uuidsentinel

from blazar import context
from blazar import tests


class TestBlazarContext(tests.TestCase):

    def test_to_dict(self):
        ctx = context.BlazarContext(
            user_id=111, project_id=222,
            request_id='req-679033b7-1755-4929-bf85-eb3bfaef7e0b')
        expected = {
            'auth_token': None,
            'domain': None,
            'global_request_id': None,
            'is_admin': False,
            'is_admin_project': True,
            'project': 222,
            'project_domain': None,
            'project_id': 222,
            'project_name': None,
            'read_only': False,
            'request_id': 'req-679033b7-1755-4929-bf85-eb3bfaef7e0b',
            'resource_uuid': None,
            'roles': [],
            'service_catalog': [],
            'show_deleted': False,
            'system_scope': None,
            'tenant': 222,
            'user': 111,
            'user_domain': None,
            'user_id': 111,
            'user_identity': '111 222 - - -',
            'user_name': None}
        self.assertEqual(expected, ctx.to_dict())

    def test_elevated_empty(self):
        ctx = context.BlazarContext.elevated()
        self.assertTrue(ctx.is_admin)

    def test_service_catalog_default(self):
        ctxt = context.BlazarContext(user_id=uuidsentinel.user_id,
                                     project_id=uuidsentinel.project_id)
        self.assertEqual([], ctxt.service_catalog)

        ctxt = context.BlazarContext(user_id=uuidsentinel.user_id,
                                     project_id=uuidsentinel.project_id,
                                     service_catalog=[])
        self.assertEqual([], ctxt.service_catalog)

        ctxt = context.BlazarContext(user_id=uuidsentinel.user_id,
                                     project_id=uuidsentinel.project_id,
                                     service_catalog=None)
        self.assertEqual([], ctxt.service_catalog)

    def test_blazar_context_elevated(self):
        user_context = context.BlazarContext(
            user_id=uuidsentinel.user_id,
            project_id=uuidsentinel.project_id, is_admin=False)
        self.assertFalse(user_context.is_admin)

        admin_context = user_context.elevated()
        self.assertFalse(user_context.is_admin)
        self.assertTrue(admin_context.is_admin)
        self.assertNotIn('admin', user_context.roles)
        self.assertIn('admin', admin_context.roles)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

from oslo_context import context


class BlazarContext(context.RequestContext):

    _context_stack = threading.local()

    def __init__(self, user_id=None, project_id=None, project_name=None,
                 service_catalog=None, user_name=None, **kwargs):
        # NOTE(neha-alhat): During serializing/deserializing context object
        # over the RPC layer, below extra parameters which are passed by
        # `oslo.messaging` are popped as these parameters are not required.
        kwargs.pop('client_timeout', None)
        kwargs.pop('user_identity', None)
        kwargs.pop('project', None)

        if user_id:
            kwargs['user_id'] = user_id
        if project_id:
            kwargs['project_id'] = project_id

        super(BlazarContext, self).__init__(**kwargs)

        self.project_name = project_name
        self.user_name = user_name
        self.service_catalog = service_catalog or []

        if self.is_admin and 'admin' not in self.roles:
            self.roles.append('admin')

    def __enter__(self):
        try:
            stack = self._context_stack.stack
        except AttributeError:
            stack = []
            self._context_stack.stack = stack
        stack.append(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        res = self._context_stack.stack.pop()
        assert res is self, "self should be the top element of the stack"

    @classmethod
    def current(cls):
        try:
            return cls._context_stack.stack[-1]
        except (AttributeError, IndexError):
            raise RuntimeError("Context isn't available here")

    # NOTE(yorik-sar): as long as oslo.rpc requires this
    def to_dict(self):
        result = super(BlazarContext, self).to_dict()
        result['user_id'] = self.user_id
        result['user_name'] = self.user_name
        result['project_id'] = self.project_id
        result['project_name'] = self.project_name
        result['service_catalog'] = self.service_catalog
        return result

    @classmethod
    def elevated(cls):
        try:
            ctx = cls.current()
        except RuntimeError:
            ctx = None
        return cls(ctx, is_admin=True)


def current():
    return BlazarContext.current()


def elevated():
    return BlazarContext.elevated()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netaddr

from blazar.manager import exceptions


def get_os_auth_host(conf):
    """Description

    Returns os_auth_host from conf, surrounded by brackets if IPv6.
    """
    os_auth_host = conf.os_auth_host
    if netaddr.valid_ipv6(os_auth_host, netaddr.core.INET_PTON):
        os_auth_host = "[%s]" % os_auth_host
    return os_auth_host


def url_for(service_catalog, service_type, admin=False,
            endpoint_interface=None,
            os_region_name=None):
    """Description

    Gets url of the service to communicate through.
    service_catalog - dict contains info about specific OpenStack service
    service_type - OpenStack service type specification
    """
    if not endpoint_interface:
        endpoint_interface = 'public'
    if admin:
        endpoint_interface = 'admin'

    service = None
    for srv in service_catalog:
        if srv['type'] == service_type:
            service = srv

    if service:
        try:
            endpoints = service['endpoints']
        except KeyError:
            raise exceptions.EndpointsNotFound(
                "No endpoints for %s" % service['type'])
        if os_region_name:
            endpoints = [e for e in endpoints if e['region'] == os_region_name]
            if not endpoints:
                raise exceptions.EndpointsNotFound("No endpoints for %s in "
                                                   "region %s" %
                                                   (service['type'],
                                                    os_region_name))
        try:
            # if Keystone API v3 endpoints returned
            endpoint = [e for e in endpoints
                        if e['interface'] == endpoint_interface][0]
            return endpoint['url']
        except KeyError:
            # otherwise
            return endpoints[0]['%sURL' % endpoint_interface]
    else:
        raise exceptions.ServiceNotFound(
            'Service "%s" not found' % service_type)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from oslo_config import cfg

from blazar import context
from blazar.utils.openstack import base
from blazar.utils.openstack import keystone

CONF = cfg.CONF


def create_trust():
    """Creates trust via Keystone API v3 to use in plugins."""
    client = keystone.BlazarKeystoneClient()

    trustee_id = keystone.BlazarKeystoneClient(
        username=CONF.os_admin_username,
        password=CONF.os_admin_password,
        tenant_name=CONF.os_admin_project_name).user_id

    ctx = context.current()
    trust = client.trusts.create(trustor_user=ctx.user_id,
                                 trustee_user=trustee_id,
                                 impersonation=False,
                                 role_names=ctx.roles,
                                 project=ctx.project_id)
    return trust


def delete_trust(lease):
    """Deletes trust for the specified lease."""
    if lease.trust_id:
        client = keystone.BlazarKeystoneClient(trust_id=lease.trust_id)
        client.trusts.delete(lease.trust_id)


def create_ctx_from_trust(trust_id):
    """Return context built from given trust."""
    ctx = context.current()

    ctx = context.BlazarContext(
        user_name=CONF.os_admin_username,
        project_name=CONF.os_admin_project_name,
        request_id=ctx.request_id,
        global_request_id=ctx.global_request_id
    )
    auth_url = "%s://%s:%s" % (CONF.os_auth_protocol,
                               base.get_os_auth_host(CONF),
                               CONF.os_auth_port)
    if CONF.os_auth_prefix:
        auth_url += "/%s" % CONF.os_auth_prefix
    client = keystone.BlazarKeystoneClient(
        password=CONF.os_admin_password,
        trust_id=trust_id,
        auth_url=auth_url,
        ctx=ctx,
    )

    # use 'with ctx' statement in the place you need context from trust
    return context.BlazarContext(
        user_name=ctx.user_name,
        project_name=ctx.project_name,
        auth_token=client.auth_token,
        service_catalog=client.service_catalog.catalog['catalog'],
        project_id=client.tenant_id,
        request_id=ctx.request_id,
        global_request_id=ctx.global_request_id
    )


def use_trust_auth():
    """Decorator creates a keystone trust

    This decorator creates a keystone trust, and adds the trust_id to the
    parameter of the decorated method.
    """
    def decorator(func):

        @functools.wraps(func)
        def wrapped(self, to_update):
            if to_update is not None:
                trust = create_trust()
                if isinstance(to_update, dict):
                    to_update.update({'trust_id': trust.id})
                elif isinstance(to_update, object):
                    setattr(to_update, 'trust_id', trust.id)
            return func(self, to_update)

        return wrapped
    return decorator



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from keystoneclient import client as keystone_client
from keystoneclient import exceptions as keystone_exception
from oslo_config import cfg

from blazar import context
from blazar import exceptions
from blazar.manager import exceptions as manager_exceptions
from blazar.utils.openstack import base


opts = [
    cfg.StrOpt('identity_service',
               default='identity',
               help='Identity service to use.'),
    cfg.StrOpt('os_region_name',
               default=None,
               help="""
Region name of this node. This is used when picking the URL in the service
catalog.

Possible values:

* Any string representing region name
""")
]

keystone_opts = [
    cfg.StrOpt('keystone_client_version',
               default='3',
               help='Keystoneclient version'),
]

CONF = cfg.CONF
CONF.register_cli_opts(opts)
CONF.register_opts(keystone_opts)


class BlazarKeystoneClient(object):
    def __init__(self, **kwargs):
        """Description

        Return Keystone client for defined in 'identity_service' conf.
        NOTE: We will use tenant_name until we start using keystone V3
        client for all our needs.

        :param version: service client version which we will use
        :type version: str

        :param username: username
        :type username: str

        :param password: password
        :type password: str

        :param tenant_name: tenant_name
        :type tenant_name: str

        :param auth_url: auth_url
        :type auth_url: str

        :param ctx: blazar context object
        :type ctx: dict

        :param auth_url: keystone auth url
        :type auth_url: string

        :param endpoint: keystone management (endpoint) url
        :type endpoint: string

        :param trust_id: keystone trust ID
        :type trust_id: string

        :param token: user token to use for authentication
        :type token: string
        """

        ctx = kwargs.pop('ctx', None)
        if ctx is None:
            try:
                ctx = context.current()
            except RuntimeError:
                pass

        kwargs.setdefault('version', cfg.CONF.keystone_client_version)
        if ctx is not None:
            kwargs.setdefault('username', ctx.user_name)
            kwargs.setdefault('tenant_name', ctx.project_name)
            kwargs.setdefault('global_request_id', ctx.global_request_id)
            if not kwargs.get('auth_url'):
                kwargs['auth_url'] = base.url_for(
                    ctx.service_catalog, CONF.identity_service,
                    os_region_name=CONF.os_region_name)
            if not kwargs.get('trust_id'):
                try:
                    kwargs.setdefault('endpoint', base.url_for(
                        ctx.service_catalog, CONF.identity_service,
                        endpoint_interface='admin',
                        os_region_name=CONF.os_region_name))
                except AttributeError:
                    raise manager_exceptions.NoManagementUrl()
            if not kwargs.get('password'):
                kwargs.setdefault('token', ctx.auth_token)

        # NOTE(dbelova): we need this checking to support current
        # keystoneclient: token can only be scoped now to either
        # a trust or project, not both.
        if kwargs.get('trust_id') and kwargs.get('tenant_name'):
            kwargs.pop('tenant_name')

        try:
            # NOTE(n.s.): we shall remove this try: except: clause when
            # https://review.openstack.org/#/c/66494/ will be merged
            self.keystone = keystone_client.Client(**kwargs)
            self.keystone.session.auth = self.keystone
            auth_url = self.complement_auth_url(kwargs.get('auth_url', None),
                                                kwargs.get('version', None))
            self.keystone.authenticate(auth_url=auth_url)
        except AttributeError:
            raise manager_exceptions.WrongClientVersion()

        self.exceptions = keystone_exception

    def complement_auth_url(self, auth_url, version):
        """Return auth_url with api version.

        This method checks whether auth_url contains api version info.
        If api version info is not contained in auth_url, this method
        complements version info to auth_url. This method only support
        complementing auth_url to v3 api url since we use keystone v3
        api to treat trusts.
        """

        # Create api version from major number of keystoneclient version.
        # keystoneclient version can take forms of "3", "v3" or "3.0" and
        # so this method convert them to form of "v3" for keystone api
        # version.
        api_version = version
        if isinstance(api_version, str):
            api_version = api_version.lstrip('v')

        api_version = int(float(api_version))

        if api_version != 3:
            raise exceptions.UnsupportedAPIVersion(version=api_version)

        if re.search(r'/v2.0/{,1}$', auth_url) is not None:
            raise exceptions.UnsupportedAPIVersion(version='v2.0')
        elif re.search(r'/v3/{,1}$', auth_url) is None:
            complemented_url = auth_url.rstrip('/') + '/v' + str(api_version)
        else:
            return auth_url

        return complemented_url

    def __getattr__(self, name):
        func = getattr(self.keystone, name)
        return func



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_serialization import jsonutils

from blazar import context
from blazar import exceptions


def ctx_from_headers(headers):
    try:
        service_catalog = jsonutils.loads(headers['X-Service-Catalog'])
    except KeyError:
        raise exceptions.ServiceCatalogNotFound()
    except TypeError:
        raise exceptions.WrongFormat()

    kwargs = {"user_id": headers['X-User-Id'],
              "project_id": headers['X-Project-Id'],
              "auth_token": headers['X-Auth-Token'],
              "service_catalog": service_catalog,
              "user_name": headers['X-User-Name'],
              "project_name": headers['X-Project-Name'],
              "roles": list(
                  map(str.strip, headers['X-Roles'].split(',')))}

    # For v1 only, request_id and global_request_id will be available.
    if headers.environ['PATH_INFO'].startswith('/v1'):
        kwargs['request_id'] = headers.environ['openstack.request_id']
        kwargs['global_request_id'] = headers.environ.get(
            'openstack.global_request_id')
    return context.BlazarContext(**kwargs)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keystoneclient import client as keystone_client

from blazar import context
from blazar import exceptions
from blazar import tests
from blazar.utils.openstack import base
from blazar.utils.openstack import keystone


class TestCKClient(tests.TestCase):

    def setUp(self):
        super(TestCKClient, self).setUp()

        self.keystone = keystone
        self.context = context
        self.k_client = keystone_client
        self.base = base

        self.ctx = self.patch(self.context, 'current')
        self.client = self.patch(self.k_client, 'Client')
        self.patch(self.base, 'url_for').return_value = 'http://fake.com/'

        self.version = '3'
        self.username = 'fake_user'
        self.token = 'fake_token'
        self.password = 'fake_pass'
        self.tenant_name = 'fake_project'
        self.auth_url = 'fake_url'
        self.trust_id = 'fake_trust'

    def test_client_from_kwargs(self):

        self.ctx.side_effect = RuntimeError

        self.keystone.BlazarKeystoneClient(version=self.version,
                                           username=self.username,
                                           password=self.password,
                                           tenant_name=self.tenant_name,
                                           trust_id=self.trust_id,
                                           auth_url=self.auth_url)

        self.client.assert_called_once_with(version=self.version,
                                            trust_id=self.trust_id,
                                            username=self.username,
                                            password=self.password,
                                            auth_url=self.auth_url)

    def test_client_from_kwargs_and_ctx(self):

        self.keystone.BlazarKeystoneClient(version=self.version,
                                           username=self.username,
                                           password=self.password,
                                           tenant_name=self.tenant_name,
                                           auth_url=self.auth_url)

        self.client.assert_called_once_with(version=self.version,
                                            tenant_name=self.tenant_name,
                                            endpoint='http://fake.com/',
                                            username=self.username,
                                            password=self.password,
                                            auth_url=self.auth_url,
                                            global_request_id=self.
                                            context.current().
                                            global_request_id)

    def test_client_from_ctx(self):

        self.keystone.BlazarKeystoneClient()

        self.client.assert_called_once_with(
            version='3',
            username=self.ctx().user_name,
            token=self.ctx().auth_token,
            tenant_name=self.ctx().project_name,
            auth_url='http://fake.com/',
            endpoint='http://fake.com/',
            global_request_id=self.context.current().global_request_id)

    def test_complement_auth_url_supported_api_version(self):
        bkc = self.keystone.BlazarKeystoneClient()
        expected_url = self.auth_url + '/v3'
        returned_url = bkc.complement_auth_url(auth_url=self.auth_url,
                                               version='v3')
        self.assertEqual(expected_url, returned_url)

    def test_complement_auth_url_unsupported_api_version(self):
        bkc = self.keystone.BlazarKeystoneClient()
        kwargs = {'auth_url': self.auth_url,
                  'version': 2.0}
        self.assertRaises(exceptions.UnsupportedAPIVersion,
                          bkc.complement_auth_url,
                          **kwargs)

    def test_complement_auth_url_valid_url(self):
        bkc = self.keystone.BlazarKeystoneClient()
        auth_url = self.auth_url + '/v3'
        returned_url = bkc.complement_auth_url(auth_url=auth_url,
                                               version='v3')
        self.assertEqual(auth_url, returned_url)

    def test_complement_auth_url_invalid_url(self):
        bkc = self.keystone.BlazarKeystoneClient()
        auth_url = self.auth_url + '/v2.0'
        kwargs = {'auth_url': auth_url,
                  'version': 'v3'}
        self.assertRaises(exceptions.UnsupportedAPIVersion,
                          bkc.complement_auth_url,
                          **kwargs)

    def test_getattr(self):
        # TODO(n.s.): Will be done as soon as pypi package will be updated
        pass
